"Parse numpy-style docstrings"

"""
Based on code from numpy, which is:
Copyright (c) 2005-2022, NumPy Developers.
All rights reserved.

Redistribution and use in source and binary forms, with or without
modification, are permitted provided that the following conditions are
met:

    * Redistributions of source code must retain the above copyright
       notice, this list of conditions and the following disclaimer.

    * Redistributions in binary form must reproduce the above
       copyright notice, this list of conditions and the following
       disclaimer in the documentation and/or other materials provided
       with the distribution.

    * Neither the name of the NumPy Developers nor the names of any
       contributors may be used to endorse or promote products derived
       from this software without specific prior written permission. """

import textwrap, re, copy
from warnings import warn
from collections import namedtuple
from collections.abc import Mapping

__all__ = ['Parameter', 'NumpyDocString', 'dedent_lines']

Parameter = namedtuple('Parameter', ['name', 'type', 'desc'])

def strip_blank_lines(l):
    "Remove leading and trailing blank lines from a list of lines"
    while l and not l[0].strip(): del l[0]
    while l and not l[-1].strip(): del l[-1]
    return l


class Reader:
    """A line-based string reader."""
    def __init__(self, data):
        if isinstance(data, list): self._str = data
        else: self._str = data.split('\n')
        self.reset()

    def __getitem__(self, n): return self._str[n]
    def reset(self): self._l = 0  # current line nr

    def read(self):
        if not self.eof():
            out = self[self._l]
            self._l += 1
            return out
        else: return ''

    def seek_next_non_empty_line(self):
        for l in self[self._l:]:
            if l.strip(): break
            else: self._l += 1

    def eof(self): return self._l >= len(self._str)

    def read_to_condition(self, condition_func):
        start = self._l
        for line in self[start:]:
            if condition_func(line): return self[start:self._l]
            self._l += 1
            if self.eof(): return self[start:self._l+1]
        return []

    def read_to_next_empty_line(self):
        self.seek_next_non_empty_line()
        def is_empty(line): return not line.strip()
        return self.read_to_condition(is_empty)

    def read_to_next_unindented_line(self):
        def is_unindented(line): return (line.strip() and (len(line.lstrip()) == len(line)))
        return self.read_to_condition(is_unindented)

    def peek(self, n=0):
        if self._l + n < len(self._str): return self[self._l + n]
        else: return ''

    def is_empty(self): return not ''.join(self._str).strip()


class ParseError(Exception):
    def __str__(self):
        message = self.args[0]
        if hasattr(self, 'docstring'): message = f"{message} in {self.docstring!r}"
        return message


class NumpyDocString(Mapping):
    """Parses a numpydoc string to an abstract representation """
    sections = { 'Summary': [''], 'Extended': [], 'Parameters': [], 'Returns': [] }

    def __init__(self, docstring, config=None):
        docstring = textwrap.dedent(docstring).split('\n')
        self._doc = Reader(docstring)
        self._parsed_data = copy.deepcopy(self.sections)
        self._parse()
        self['Parameters'] = {o.name:o for o in self['Parameters']}
        if self['Returns']: self['Returns'] = self['Returns'][0]
        self['Summary'] = dedent_lines(self['Summary'], split=False)
        self['Extended'] = dedent_lines(self['Extended'], split=False)

    def __iter__(self): return iter(self._parsed_data)
    def __len__(self): return len(self._parsed_data)
    def __getitem__(self, key): return self._parsed_data[key]

    def __setitem__(self, key, val):
        if key not in self._parsed_data: self._error_location(f"Unknown section {key}", error=False)
        else: self._parsed_data[key] = val

    def _is_at_section(self):
        self._doc.seek_next_non_empty_line()
        if self._doc.eof(): return False
        l1 = self._doc.peek().strip()  # e.g. Parameters
        l2 = self._doc.peek(1).strip()  # ---------- or ==========
        if len(l2) >= 3 and (set(l2) in ({'-'}, {'='}) ) and len(l2) != len(l1):
            snip = '\n'.join(self._doc._str[:2])+'...'
            self._error_location("potentially wrong underline length... \n%s \n%s in \n%s" % (l1, l2, snip), error=False)
        return l2.startswith('-'*len(l1)) or l2.startswith('='*len(l1))

    def _strip(self, doc):
        i = 0
        j = 0
        for i, line in enumerate(doc):
            if line.strip(): break
        for j, line in enumerate(doc[::-1]):
            if line.strip(): break
        return doc[i:len(doc)-j]

    def _read_to_next_section(self):
        section = self._doc.read_to_next_empty_line()

        while not self._is_at_section() and not self._doc.eof():
            if not self._doc.peek(-1).strip(): section += ['']
            section += self._doc.read_to_next_empty_line()
        return section

    def _read_sections(self):
        while not self._doc.eof():
            data = self._read_to_next_section()
            name = data[0].strip()

            if name.startswith('..'): yield name, data[1:]
            elif len(data) < 2: yield StopIteration
            else: yield name, self._strip(data[2:])

    def _parse_param_list(self, content, single_element_is_type=False):
        content = dedent_lines(content)
        r = Reader(content)
        params = []
        while not r.eof():
            header = r.read().strip()
            if ' :' in header:
                arg_name, arg_type = header.split(' :', maxsplit=1)
                arg_name, arg_type = arg_name.strip(), arg_type.strip()
            else:
                if single_element_is_type: arg_name, arg_type = '', header
                else: arg_name, arg_type = header, ''

            desc = r.read_to_next_unindented_line()
            desc = dedent_lines(desc)
            desc = strip_blank_lines(desc)
            params.append(Parameter(arg_name, arg_type, desc))
        return params

    def _parse_summary(self):
        """Grab signature (if given) and summary"""
        if self._is_at_section(): return

        # If several signatures present, take the last one
        while True:
            summary = self._doc.read_to_next_empty_line()
            summary_str = " ".join([s.strip() for s in summary]).strip()
            compiled = re.compile(r'^([\w., ]+=)?\s*[\w\.]+\(.*\)$')
            if compiled.match(summary_str) and not self._is_at_section(): continue
            break

        if summary is not None: self['Summary'] = summary
        if not self._is_at_section(): self['Extended'] = self._read_to_next_section()

    def _parse(self):
        self._doc.reset()
        self._parse_summary()

        sections = list(self._read_sections())
        section_names = {section for section, content in sections}

        has_returns = 'Returns' in section_names
        has_yields = 'Yields' in section_names
        # We could do more tests, but we are not. Arbitrarily.
        if has_returns and has_yields:
            msg = 'Docstring contains both a Returns and Yields section.'
            raise ValueError(msg)
        if not has_yields and 'Receives' in section_names:
            msg = 'Docstring contains a Receives section but not Yields.'
            raise ValueError(msg)

        for (section, content) in sections:
            if not section.startswith('..'):
                section = (s.capitalize() for s in section.split(' '))
                section = ' '.join(section)
                if self.get(section):
                    self._error_location("The section %s appears twice in  %s" % (section, '\n'.join(self._doc._str)))

            if section in ('Parameters', 'Other Parameters', 'Attributes', 'Methods'):
                self[section] = self._parse_param_list(content)
            elif section in ('Returns', 'Yields', 'Raises', 'Warns', 'Receives'):
                self[section] = self._parse_param_list( content, single_element_is_type=True)
            else: self[section] = content

    @property
    def _obj(self):
        if hasattr(self, '_cls'): return self._cls
        elif hasattr(self, '_f'): return self._f
        return None

    def _error_location(self, msg, error=True):
        if error: raise ValueError(msg)
        else: warn(msg)


def dedent_lines(lines, split=True):
    """Deindent a list of lines maximally"""
    res = textwrap.dedent("\n".join(lines))
    if split: res = res.split("\n")
    return res




from functools import wraps
import shutil

__all__ = ['copymode', 'copystat', 'copy', 'copy2', 'move', 'copytree', 'rmtree', 'disk_usage', 'chown', 'rmtree']

def str_src_dest(f):
    @wraps(f)
    def _f(src, dst, *args, **kwargs): return f(str(src), str(dst), *args, **kwargs)
    return _f

def str_path(f):
    @wraps(f)
    def _f(path, *args, **kwargs): return f(str(path), *args, **kwargs)
    return _f

src_dests = ['copymode', 'copystat', 'copy', 'copy2', 'move', 'copytree']
for o in src_dests: globals()[o] = str_src_dest(getattr(shutil,o))

paths = ['rmtree', 'disk_usage', 'chown', 'rmtree']
for o in paths: globals()[o] = str_path(getattr(shutil,o))




# AUTOGENERATED! DO NOT EDIT! File to edit: ../nbs/00_test.ipynb.

# %% auto 0
__all__ = ['TEST_IMAGE', 'TEST_IMAGE_BW', 'exception', 'test_fail', 'test', 'nequals', 'test_eq', 'test_eq_type', 'test_ne',
           'is_close', 'test_close', 'test_is', 'test_shuffled', 'test_stdout', 'test_warns', 'test_fig_exists',
           'ExceptionExpected']

# %% ../nbs/00_test.ipynb 1
from .imports import *
from collections import Counter
from contextlib import redirect_stdout

# %% ../nbs/00_test.ipynb 6
def test_fail(f, msg='', contains='', args=None, kwargs=None):
    "Fails with `msg` unless `f()` raises an exception and (optionally) has `contains` in `e.args`"
    args, kwargs = args or [], kwargs or {}
    try: f(*args, **kwargs)
    except Exception as e:
        assert not contains or contains in str(e)
        return
    assert False,f"Expected exception but none raised. {msg}"

# %% ../nbs/00_test.ipynb 10
def test(a, b, cmp, cname=None):
    "`assert` that `cmp(a,b)`; display inputs and `cname or cmp.__name__` if it fails"
    if cname is None: cname=cmp.__name__
    assert cmp(a,b),f"{cname}:\n{a}\n{b}"

# %% ../nbs/00_test.ipynb 16
def nequals(a,b):
    "Compares `a` and `b` for `not equals`"
    return not equals(a,b)

# %% ../nbs/00_test.ipynb 20
def test_eq(a,b):
    "`test` that `a==b`"
    test(a,b,equals, cname='==')

# %% ../nbs/00_test.ipynb 25
def test_eq_type(a,b):
    "`test` that `a==b` and are same type"
    test_eq(a,b)
    test_eq(type(a),type(b))
    if isinstance(a,(list,tuple)): test_eq(map(type,a),map(type,b))

# %% ../nbs/00_test.ipynb 27
def test_ne(a,b):
    "`test` that `a!=b`"
    test(a,b,nequals,'!=')

# %% ../nbs/00_test.ipynb 29
def is_close(a,b,eps=1e-5):
    "Is `a` within `eps` of `b`"
    if hasattr(a, '__array__') or hasattr(b,'__array__'):
        return (abs(a-b)<eps).all()
    if isinstance(a, (Iterable,Generator)) or isinstance(b, (Iterable,Generator)):
        return all(abs(a_-b_)<eps for a_,b_ in zip(a,b))
    return abs(a-b)<eps

# %% ../nbs/00_test.ipynb 30
def test_close(a,b,eps=1e-5):
    "`test` that `a` is within `eps` of `b`"
    test(a,b,partial(is_close,eps=eps),'close')

# %% ../nbs/00_test.ipynb 32
def test_is(a,b):
    "`test` that `a is b`"
    test(a,b,operator.is_, 'is')

# %% ../nbs/00_test.ipynb 34
def test_shuffled(a,b):
    "`test` that `a` and `b` are shuffled versions of the same sequence of items"
    test_ne(a, b)
    test_eq(Counter(a), Counter(b))

# %% ../nbs/00_test.ipynb 38
def test_stdout(f, exp, regex=False):
    "Test that `f` prints `exp` to stdout, optionally checking as `regex`"
    s = io.StringIO()
    with redirect_stdout(s): f()
    if regex: assert re.search(exp, s.getvalue()) is not None, f"regex '{exp}' did not not match stdout '{s.getvalue()}'"
    else: test_eq(s.getvalue(), f'{exp}\n' if len(exp) > 0 else '')

# %% ../nbs/00_test.ipynb 40
def test_warns(f, show=False):
    with warnings.catch_warnings(record=True) as w:
        f()
        assert w, "No warnings raised"
        if show:
            for e in w: print(f"{e.category}: {e.message}")

# %% ../nbs/00_test.ipynb 43
TEST_IMAGE = 'images/puppy.jpg'

# %% ../nbs/00_test.ipynb 45
TEST_IMAGE_BW = 'images/mnist3.png'

# %% ../nbs/00_test.ipynb 47
def test_fig_exists(ax):
    "Test there is a figure displayed in `ax`"
    assert ax and len(ax.figure.canvas.tostring_argb())

# %% ../nbs/00_test.ipynb 50
class ExceptionExpected:
    "Context manager that tests if an exception is raised"
    def __init__(self, ex=Exception, regex=''): self.ex,self.regex = ex,regex
    def __enter__(self): pass
    def __exit__(self, type, value, traceback):
        if not isinstance(value, self.ex) or (self.regex and not re.search(self.regex, f'{value.args}')):
            raise TypeError(f"Expected {self.ex.__name__}({self.regex}) not raised.")
        return True

# %% ../nbs/00_test.ipynb 53
exception = ExceptionExpected()
